@@ -125,11 +125,20 @@ int inet_ptom (int af, const char *src, unsigned int *dst)
  */
 const char *inet_mtop (int af, unsigned int src, char *dst, socklen_t len)
 {
-    struct in_addr addr;
-    
-    inet_mton(AF_INET, src, &addr);
-    
-    return inet_ntop (af, &addr, dst, len);
+    if (af == AF_INET) {
+        struct in_addr addr;
+
+        inet_mton(af, src, &addr);
+        return inet_ntop(af, &addr, dst, len);
+    } else if (af == AF_INET6) {
+        struct in6_addr *addr6;
+
+        inet_mton(af, src, &addr6);
+        return inet_ntop(af, &addr6, dst, len);
+    } else {
+        di_warning("Unknown address family passed to inet_mtop: %i", af);
+        return NULL;
+    }
 }
 
 /* convert a mask length (eg 24) in +src+ into the struct in_addr it corresponds
@@ -1356,11 +1365,12 @@ void reap_old_files (void)
 /* Convert a space-separated list of nameservers in a single string (as might
  * be entered into, say, debconf), and store them in the interface.
  */
-void netcfg_nameservers_to_array(const char *nameservers, struct netcfg_interface *interface)
+int netcfg_nameservers_to_array(const char *nameservers, struct netcfg_interface *interface)
 {
     char *save, *ptr, *ns;
     unsigned int i;
     union inX_addr addr;
+    int ok = 1;
 
     if (nameservers) {
         save = ptr = strdup(nameservers);
@@ -1383,7 +1393,8 @@ void netcfg_nameservers_to_array(const char *nameservers, struct netcfg_interfac
                         af = AF_INET6;
                     } else {
                         af = -1;
-                        fprintf(stderr, "Failed to parse %s as an IP address", ns);
+                        fprintf(stderr, "Failed to parse %s as an IP address\n", ns);
+                        ok = 0;
                     }
                 }
 
@@ -1402,12 +1413,14 @@ void netcfg_nameservers_to_array(const char *nameservers, struct netcfg_interfac
         /* Empty out all the nameserver strings */
         for (i = 0; i < NETCFG_NAMESERVERS_MAX; i++) *(interface->nameservers[i]) = '\0';
     }
+    return ok;
 }
 
-int netcfg_get_nameservers (struct debconfclient *client, char **nameservers, char *default_nameservers)
+int netcfg_get_nameservers (struct debconfclient *client, char **nameservers, char *default_nameservers,
+                            struct netcfg_interface *interface)
 {
     char *ptr;
-    int ret;
+    int ret, ok = 0;
 
     debconf_get(client,"netcfg/get_nameservers");
     if (*nameservers)
@@ -1420,14 +1433,24 @@ int netcfg_get_nameservers (struct debconfclient *client, char **nameservers, ch
         ptr = "";
     debconf_set(client, "netcfg/get_nameservers", ptr);
 
-    debconf_input(client, "critical", "netcfg/get_nameservers");
-    ret = debconf_go(client);
+    while (!ok) {
+        debconf_input(client, "critical", "netcfg/get_nameservers");
+        ret = debconf_go(client);
 
-    if (ret)
-        return ret;
+        if (ret)
+            return ret;
 
-    debconf_get(client, "netcfg/get_nameservers");
-    ptr = client->value;
+        debconf_get(client, "netcfg/get_nameservers");
+        ptr = client->value;
+        ok = netcfg_nameservers_to_array(ptr, interface);
+
+        if (!ok) {
+            debconf_capb(client);
+            debconf_input (client, "critical", "netcfg/bad_ipaddress");
+            debconf_go (client);
+            debconf_capb(client, "backup");
+        }
+    }
 
     if (*nameservers)
         free(*nameservers);
@@ -1661,8 +1684,22 @@ int netcfg_gateway_reachable(const struct netcfg_interface *interface)
     inet_pton(interface->address_family, interface->gateway, &gw_addr);
 
     if (interface->address_family == AF_INET) {
+        if (interface->masklen == 32) {
+            /* Special case for single-address networks.  We'll handle
+             * routing manually in this case.
+             */
+            return 1;
+        }
+
         return (gw_addr.in4.s_addr && ((gw_addr.in4.s_addr & mask.in4.s_addr) == net.in4.s_addr));
     } else if (interface->address_family == AF_INET6) {
+        if (interface->masklen == 128) {
+            /* Special case for single-address networks.  We'll handle
+             * routing manually in this case.
+             */
+            return 1;
+        }
+
         if ((ntohs(gw_addr.in6.s6_addr16[0]) & 0xffc0) == (0xfe80 & 0xffc0)) {
             return 1;
         }
